"""Utility classes"""

import string
import os
import ihm
import datetime


class _AsymIDs:
    """Map indices to multi-character asym (chain) IDs.
       We label the first 26 chains A-Z, then we move to two-letter
       chain IDs: AA through AZ, then BA through BZ, through to ZZ.
       This continues with longer chain IDs."""
    def __getitem__(self, ind):
        chars = string.ascii_uppercase
        lc = len(chars)
        ids = []
        while ind >= lc:
            ids.append(chars[ind % lc])
            ind = ind // lc - 1
        ids.append(chars[ind])
        return "".join(reversed(ids))


def _remove_id(obj, attr='_id'):
    """Remove any unique ID from obj"""
    if hasattr(obj, attr):
        delattr(obj, attr)


def _assign_id(obj, seen_objs, obj_by_id, attr='_id', seen_obj=None,
               by_id_obj=None):
    """Assign a unique ID to obj, and track all ids in obj_by_id."""
    if seen_obj is None:
        seen_obj = obj
    if by_id_obj is None:
        by_id_obj = obj
    if seen_obj not in seen_objs:
        if not hasattr(obj, attr):
            obj_by_id.append(by_id_obj)
            setattr(obj, attr, len(obj_by_id))
        seen_objs[seen_obj] = getattr(obj, attr)
    else:
        setattr(obj, attr, seen_objs[seen_obj])


def _get_relative_path(reference, path):
    """Return `path` interpreted relative to `reference`"""
    if os.path.isabs(path):
        return path
    else:
        return os.path.join(os.path.dirname(reference), path)


def _text_choice_property(attr, choices, doc=None):
    """Like `property` but requires that the value be one of the set choices"""
    schoices = frozenset(choices)

    def getfunc(obj):
        return getattr(obj, "_" + attr)

    def setfunc(obj, val):
        if val is not None and val is not ihm.unknown and val not in schoices:
            raise ValueError(
                "Invalid choice %s for %s; valid values are %s, "
                "None, ihm.unknown"
                % (repr(val), attr, ", ".join(repr(x) for x in choices)))
        setattr(obj, "_" + attr, val)

    return property(getfunc, setfunc, doc=doc)


def _check_residue_range(seq_id_range, entity):
    """Make sure that a residue range is not out of range of its Entity"""
    if not entity or not entity._range_check:
        return
    if seq_id_range[1] < seq_id_range[0]:
        raise ValueError("Range %d-%d is invalid; end is before start"
                         % seq_id_range)
    if (seq_id_range[1] > len(entity.sequence)
            or seq_id_range[0] < 1):
        raise IndexError("Range %d-%d out of range for %s (1-%d)"
                         % (seq_id_range[0], seq_id_range[1],
                            entity, len(entity.sequence)))


def _check_residue(r):
    """Make sure that a residue is not out of range of its Entity"""
    if not r.entity or not r.entity._range_check:
        return
    if r.seq_id > len(r.entity.sequence) or r.seq_id < 1:
        raise IndexError("Residue %d out of range for %s (1-%d)"
                         % (r.seq_id, r.entity, len(r.entity.sequence)))


def _check_transform(t):
    if t.rot_matrix in (None, ihm.unknown):
        raise ValueError("Transformation %s is missing rotation" % t)
    if t.tr_vector in (None, ihm.unknown):
        raise ValueError("Transformation %s is missing translation" % t)


def _invert_ranges(ranges, end, start=1):
    """Given a sorted list of non-overlapping ranges, yield a new list which
       contains every range in the range start-end which was not in the
       original list.  For example, if end=4,
       [(2, 3)] -> [(1, 1), (4, 4)]"""
    for r in ranges:
        if r[0] > start:
            yield (start, r[0] - 1)
        start = r[1] + 1
    if end >= start:
        yield (start, end)


def _pred_ranges(ranges, end):
    """Given a sorted list of non-overlapping ranges, yield a new list which
       covers the range 1-end. Each element in the new list contains a new
       third bool member which is True iff the element was in the original
       list. For example, if end=4,
       [(2, 3)] -> [(1, 1, False), (2, 3, True), (4, 4, False)]"""
    start = 1
    for r in ranges:
        if r[0] > start:
            yield (start, r[0] - 1, False)
        yield (r[0], r[1], True)
        start = r[1] + 1
    if end >= start:
        yield (start, end, False)


def _combine_ranges(ranges):
    """Sort the input ranges and remove any overlaps; yield the result.
       For example, [(8, 10), (1, 2), (3, 4)] -> [(1, 4), (8, 10)]"""
    ranges = sorted(ranges)
    if not ranges:
        return
    current = ranges[0]
    for r in ranges[1:]:
        if current[1] + 1 >= r[0]:
            current = (current[0], max(r[1], current[1]))
        else:
            yield current
            current = r
    yield current


def _make_range_from_list(rr):
    """Yield a list of ranges given a sorted list of values.
       For example, [1, 2, 5, 6] -> [[1, 2], [5, 6]]"""
    if not rr:
        return
    current = [rr[0], rr[0]]
    for r in rr[1:]:
        if current[1] + 1 == r:
            current[1] = r
        else:
            yield current
            current = [r, r]
    yield current


def _get_codes(codestr):
    """Convert a one-letter-code string into a sequence of individual
       codes"""
    if codestr is None or codestr is ihm.unknown:
        return
    i = 0
    while i < len(codestr):
        # Strip out linebreaks
        if codestr[i] == '\n':
            pass
        elif codestr[i] == '(':
            end = codestr.index(')', i)
            yield codestr[i + 1:end]
            i = end
        else:
            yield codestr[i]
        i += 1


def _get_iso_date(iso_date_str):
    """Get a datetime.date obj for a string in isoformat."""
    if not iso_date_str:
        return iso_date_str
    return datetime.date(int(iso_date_str[0:4]),
                         int(iso_date_str[5:7]),
                         int(iso_date_str[8:10]))



from io import BytesIO
import urllib.error

zinc_atoms = b'_chem_comp_atom.comp_id ZN\n_chem_comp_atom.atom_id ZN\n'


def urlopen(url):
    if 'invalid' in url:
        raise urllib.error.HTTPError("404")
    return BytesIO(zinc_atoms)



"""Helper classes to provide a summary report of an :class:`ihm.System`"""

import ihm
import sys
import warnings
import collections


class MissingDataWarning(UserWarning):
    pass


class LocalFilesWarning(UserWarning):
    pass


class MissingFileWarning(UserWarning):
    pass


def _get_name(name):
    if name:
        return repr(name)
    else:
        return "(unnamed)"


class _SectionReporter:
    def __init__(self, title, fh):
        self.fh = fh
        print("\n\n# " + title, file=self.fh)

    def report(self, txt):
        print("   " + str(txt), file=self.fh)


class Reporter:
    def __init__(self, system, fh=sys.stdout):
        self.system = system
        self.fh = fh

    def report(self):
        print("Title: %s" % self.system.title, file=self.fh)
        self.report_entities()
        self.report_asyms()
        self.report_representations()
        self.report_databases()
        self.report_files()
        self.report_citations()
        self.report_software()
        self.report_protocols()
        self.report_restraints()
        self.report_models()
        self.report_ensembles()

    def _section(self, title):
        return _SectionReporter(title, self.fh)

    def report_entities(self):
        r = self._section("Entities (unique sequences)")
        asyms_for_entity = collections.defaultdict(list)
        for a in self.system.asym_units:
            asyms_for_entity[a.entity].append(a)
        for e in self.system.entities:
            asyms = asyms_for_entity[e]
            r.report("- %s (length %d, %d instances, chain IDs %s)"
                     % (e.description, len(e.sequence), len(asyms),
                        ", ".join(a.id for a in asyms)))
            if len(e.references) == 0:
                warnings.warn(
                    "No reference sequence (e.g. from UniProt) provided "
                    "for %s" % e, MissingDataWarning)
            for ref in e.references:
                r.report("  - from %s" % str(ref))

    def report_asyms(self):
        r = self._section("Asyms/chains")
        for a in self.system.asym_units:
            r.report("- %s (chain ID %s)" % (a.details, a.id))

    def report_citations(self):
        r = self._section("Publications cited")
        for c in self.system._all_citations():
            r.report('- "%s", %s, %s' % (c.title, c.journal, c.year))

    def report_software(self):
        r = self._section("Software used")
        for s in ihm._remove_identical(self.system._all_software()):
            if s.version is None:
                r.report("- %s (no version given)" % s.name)
            else:
                r.report("- %s (version %s)" % (s.name, s.version))
            if not s.citation:
                warnings.warn(
                    "No citation provided for %s" % s, MissingDataWarning)

    def report_databases(self):
        r = self._section("External databases referenced")
        for loc in ihm._remove_identical(self.system._all_locations()):
            if isinstance(loc, ihm.location.DatabaseLocation):
                r.report(" - %s accession %s"
                         % (loc.db_name, loc.access_code))

    def report_files(self):
        r = self._section("Additional files referenced")
        locs_by_repo = collections.defaultdict(list)
        for loc in ihm._remove_identical(self.system._all_locations()):
            if not isinstance(loc, ihm.location.DatabaseLocation):
                locs_by_repo[loc.repo].append(loc)
        for repo, locs in locs_by_repo.items():
            r.report("- %s" % ("DOI: " + repo.doi if repo else "Local files"))
            for loc in locs:
                r.report("  - %r, %s" % (loc.path, loc.details))
        if None in locs_by_repo:
            warnings.warn(
                "The following local files are referenced (they will need to "
                "be deposited in a database or with a DOI): %s"
                % [loc.path for loc in locs_by_repo[None]], LocalFilesWarning)

    def report_representations(self):
        r = self._section("Model representation")
        for rep in self.system._all_representations():
            if hasattr(rep, '_id'):
                r.report("- Representation %s" % rep._id)
            else:
                r.report("- Representation")
            for segment in rep:
                r.report("  - " + segment._get_report())

    def report_protocols(self):
        r = self._section("Modeling protocols")
        for prot in self.system._all_protocols():
            r.report("- " + (prot.name if prot.name else "Unnamed protocol"))
            for step in prot.steps:
                r.report("  - " + step._get_report())
            for analysis in prot.analyses:
                r.report("  - Analysis")
                for step in analysis.steps:
                    r.report("    - " + step._get_report())

    def report_restraints(self):
        r = self._section("Restraints")
        for rsr in ihm._remove_identical(self.system._all_restraints()):
            r.report("- " + rsr._get_report())

    def report_models(self):
        r = self._section("Models")
        for sg in self.system.state_groups:
            r.report("- State group")
            for state in sg:
                r.report("  - State %s" % _get_name(state.name))
                for mg in state:
                    r.report("    - Model group %s containing %d models"
                             % (_get_name(mg.name), len(mg)))

    def report_ensembles(self):
        r = self._section("Ensembles")
        for e in self.system.ensembles:
            r.report("- Ensemble %s containing %d models"
                     % (_get_name(e.name), e.num_models))
            if e.model_group is not None:
                r.report("  - From model group %s"
                         % _get_name(e.model_group.name))
            if e.precision is not None:
                r.report("  - Precision %.1f" % e.precision)
            if e.file:
                r.report("  - In external file %s" % e.file)
            if (e.model_group is not None and not e.file
                    and e.num_models > len(e.model_group)):
                warnings.warn(
                    "%s references more models (%d) than are deposited in "
                    "its model group, but does not reference an external file"
                    % (e, e.num_models), MissingFileWarning)
            for d in e.densities:
                asym = d.asym_unit
                r.report("  - Localization density for %s %d-%d"
                         % (asym.details, asym.seq_id_range[0],
                            asym.seq_id_range[1]))



#!/usr/bin/python3

"""Get the set of canonical atom names for one or more residue types,
   by querying CCD via RCSB. Used to populate the KNOWN_ATOM_NAMES variable
   in make_mmcif.py.
"""

import collections
import ihm.format
import urllib.request
import sys


class ChemCompAtomHandler:
    not_in_file = omitted = unknown = None

    def __init__(self):
        super().__init__()
        self.atoms = collections.defaultdict(set)

    def __call__(self, comp_id, atom_id):
        self.atoms[comp_id].add(atom_id)


cca = ChemCompAtomHandler()
for resname in sys.argv[1:]:
    with urllib.request.urlopen(
            'https://files.rcsb.org/pub/pdb/refdata/chem_comp/%s/%s/%s.cif'
            % (resname[-1], resname, resname)) as fh:
        c = ihm.format.CifReader(fh, category_handler={'_chem_comp_atom': cca})
        c.read_file()

print("KNOWN_ATOM_NAMES = {")
for comp_id in sorted(cca.atoms.keys()):
    print('    %r: %r,' % (comp_id, cca.atoms[comp_id]))
print("}")



import unittest
import utils
import os
from io import StringIO

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import ihm
import ihm.report
import ihm.reference
import ihm.location
import ihm.representation
import ihm.protocol
import ihm.analysis
import ihm.restraint
import ihm.geometry
import ihm.model


class Tests(unittest.TestCase):
    def test_report(self):
        """Test System.report()"""
        sio = StringIO()
        s = ihm.System(title='test system')
        s.report(sio)

    def test_entities(self):
        """Test report_entities"""
        sio = StringIO()
        s = ihm.System(title='test system')
        e = ihm.Entity("ACG")
        a = ihm.AsymUnit(e, "my asym")
        a.id = 'A'
        s.asym_units.append(a)
        s.entities.append(e)
        r = ihm.report.Reporter(s, sio)
        # Should warn about missing references
        self.assertWarns(ihm.report.MissingDataWarning, r.report_entities)
        uniprot = ihm.reference.UniProtSequence(
            db_code='testcode', accession='testacc', sequence='CCCG')
        e.references.append(uniprot)
        r.report_entities()

    def test_asyms(self):
        """Test report_asyms"""
        sio = StringIO()
        s = ihm.System(title='test system')
        e = ihm.Entity("ACG")
        s.entities.append(e)
        a = ihm.AsymUnit(e, "my asym")
        s.asym_units.append(a)
        r = ihm.report.Reporter(s, sio)
        r.report_asyms()

    def test_representations(self):
        """Test report_representations"""
        sio = StringIO()
        s = ihm.System(title='test system')
        e = ihm.Entity("ACGT")
        s.entities.append(e)
        a = ihm.AsymUnit(e, "my asym")
        s.asym_units.append(a)
        s1 = ihm.representation.ResidueSegment(
            a(1, 2), starting_model=None,
            rigid=False, primitive='sphere')
        s2 = ihm.representation.FeatureSegment(
            a(3, 4), starting_model=None,
            rigid=True, primitive='other', count=3)
        r1 = ihm.representation.Representation((s1, s2), name='foo')
        s.orphan_representations.append(r1)
        r = ihm.report.Reporter(s, sio)
        # Test report of representation without ID
        r.report_representations()
        # Test report of representation with ID
        r1._id = 42
        r.report_representations()

    def test_citations(self):
        """Test report_citations"""
        sio = StringIO()
        s = ihm.System(title='test system')
        c = ihm.Citation(pmid="foo", title="bar", journal="j", volume=1,
                         page_range=(10, 20), year=2023,
                         authors=["foo", "bar"], doi="test")
        s.citations.append(c)
        r = ihm.report.Reporter(s, sio)
        r.report_citations()

    def test_software(self):
        """Test report_software"""
        sio = StringIO()
        s = ihm.System(title='test system')
        soft = ihm.Software(name='foo', version='1.0',
                            classification='1', description='2', location='3')
        s.software.append(soft)
        r = ihm.report.Reporter(s, sio)
        # Should warn about missing citation
        self.assertWarns(ihm.report.MissingDataWarning, r.report_software)
        c = ihm.Citation(pmid="foo", title="bar", journal="j", volume=1,
                         page_range=(10, 20), year=2023,
                         authors=["foo", "bar"], doi="test")
        soft.citation = c
        soft.version = None
        r.report_software()

    def test_databases(self):
        """Test report_databases"""
        sio = StringIO()
        s = ihm.System(title='test system')
        s.locations.append(
            ihm.location.BMRBLocation('27600', version='foo', details='bar'))
        s.locations.append(
            ihm.location.FileLocation(repo='mydoi', path='a'))
        r = ihm.report.Reporter(s, sio)
        r.report_databases()

    def test_files(self):
        """Test report_files"""
        sio = StringIO()
        s = ihm.System(title='test system')
        repo = ihm.location.Repository(doi='1.2.3.4')
        s.locations.append(
            ihm.location.BMRBLocation('27600', version='foo', details='bar'))
        s.locations.append(
            ihm.location.FileLocation(repo=repo, path='a'))
        r = ihm.report.Reporter(s, sio)
        r.report_files()

    def test_files_local(self):
        """Test report_files with local files"""
        sio = StringIO()
        s = ihm.System(title='test system')
        s.locations.append(
            ihm.location.BMRBLocation('27600', version='foo', details='bar'))
        s.locations.append(
            ihm.location.FileLocation(repo=None, path='.'))
        r = ihm.report.Reporter(s, sio)
        # Should warn about local files
        self.assertWarns(ihm.report.LocalFilesWarning, r.report_files)

    def test_protocols(self):
        """Test report_protocols"""
        sio = StringIO()
        s = ihm.System(title='test system')
        prot = ihm.protocol.Protocol(name='foo')
        prot.steps.append(ihm.protocol.Step(
            assembly='foo', dataset_group='bar',
            method='baz', num_models_begin=0, num_models_end=42))
        analysis = ihm.analysis.Analysis()
        analysis.steps.append(ihm.analysis.ClusterStep(
            feature='RMSD', num_models_begin=42, num_models_end=5))
        prot.analyses.append(analysis)
        s.orphan_protocols.append(prot)
        r = ihm.report.Reporter(s, sio)
        r.report_protocols()

    def test_restraints(self):
        """Test report_restraints"""
        sio = StringIO()
        s = ihm.System(title='test system')
        dist = ihm.restraint.UpperBoundDistanceRestraint(42.0)
        geom = ihm.geometry.XAxis(name='foo', description='bar')
        rsr = ihm.restraint.GeometricRestraint(
            dataset='foo', geometric_object=geom, feature='feat',
            distance=dist)
        s.restraints.append(rsr)
        r = ihm.report.Reporter(s, sio)
        r.report_restraints()

    def test_models(self):
        """Test report_models"""
        sio = StringIO()
        s = ihm.System(title='test system')
        sg = ihm.model.StateGroup()
        state = ihm.model.State(name='foo')
        mg = ihm.model.ModelGroup(name='bar')
        m = ihm.model.Model(assembly='foo', protocol='bar',
                            representation='baz')
        mg.append(m)
        state.append(mg)
        sg.append(state)
        s.state_groups.append(sg)
        r = ihm.report.Reporter(s, sio)
        r.report_models()

    def test_ensembles(self):
        """Test report_ensembles"""
        sio = StringIO()
        s = ihm.System(title='test system')
        e = ihm.Entity("ACG")
        a = ihm.AsymUnit(e, "my asym")
        a.id = 'A'
        s.asym_units.append(a)
        s.entities.append(e)
        e1 = ihm.model.Ensemble(model_group=None, num_models=10)
        e2 = ihm.model.Ensemble(model_group=None, num_models=5, precision=1.0,
                                name='test ensemble')
        mg = ihm.model.ModelGroup(name='bar')
        e3 = ihm.model.Ensemble(model_group=mg, num_models=1, file='file')
        e3.densities.append(ihm.model.LocalizationDensity(
            file='foo', asym_unit=a(1, 2)))
        s.ensembles.extend((e1, e2, e3))
        r = ihm.report.Reporter(s, sio)
        r.report_ensembles()
        # Should warn about extra models but no external file
        e4 = ihm.model.Ensemble(model_group=mg, num_models=1)
        s.ensembles.append(e4)
        self.assertWarns(ihm.report.MissingFileWarning, r.report_ensembles)


if __name__ == '__main__':
    unittest.main()



import utils
import os
import unittest
import sys
import struct
from io import BytesIO

TOPDIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
utils.set_search_paths(TOPDIR)
import ihm.format_bcif

try:
    from ihm import _format
except ImportError:
    _format = None


# Provide dummy implementations of msgpack.unpack() and msgpack.pack() which
# just return the data unchanged. We can use these to test the Python BinaryCIF
# parser with Python objects rather than having to install msgpack and
# generate real binary files
class MockMsgPack:
    @staticmethod
    def unpack(fh, raw=False):
        return fh

    @staticmethod
    def pack(data, fh, use_bin_type=True):
        fh.data = data


class MockFh:
    pass


class GenericHandler:
    """Capture BinaryCIF data as a simple list of dicts"""
    not_in_file = None
    omitted = None
    unknown = "?"

    _keys = ('method', 'foo', 'bar', 'baz', 'pdbx_keywords', 'var1',
             'var2', 'var3', 'intkey1', 'intkey2', 'floatkey1', 'floatkey2',
             'boolkey1')
    _int_keys = frozenset(('intkey1', 'intkey2'))
    _float_keys = frozenset(('floatkey1', 'floatkey2'))
    _bool_keys = frozenset(('boolkey1',))

    def __init__(self):
        self.data = []

    def __call__(self, *args):
        d = {}
        for k, v in zip(self._keys, args):
            if v is not None:
                d[k] = v
        self.data.append(d)


def _encode_int(rows):
    d = {'data': struct.pack('%db' % len(rows), *rows),
         'encoding': [{'kind': 'ByteArray',
                       'type': ihm.format_bcif._Uint8}]}
    return d, None


def _encode_float(rows):
    d = {'data': struct.pack('<%dd' % len(rows), *rows),
         'encoding': [{'kind': 'ByteArray',
                       'type': ihm.format_bcif._Float64}]}
    return d, None


def _encode(rows):
    # Assume rows is a list of strings; make simple BinaryCIF encoding
    mask = [0] * len(rows)
    need_mask = False
    for i, row in enumerate(rows):
        if row is None:
            need_mask = True
            mask[i] = 1
        elif row == '?':
            need_mask = True
            mask[i] = 2
    if need_mask:
        rows = ['' if r == '?' or r is None else r for r in rows]
        mask = {'data': ''.join(chr(i) for i in mask).encode('ascii'),
                'encoding': [{'kind': 'ByteArray',
                              'type': ihm.format_bcif._Uint8}]}
    else:
        mask = None
    string_data = "".join(rows)

    offsets = []
    total_len = 0
    for row in rows:
        offsets.append(total_len)
        total_len += len(row)
    offsets.append(total_len)
    offsets = ''.join(chr(i) for i in offsets).encode('ascii')
    indices = ''.join(chr(i) for i in range(len(rows))).encode('ascii')
    string_array_encoding = {
        'kind': 'StringArray',
        'dataEncoding': [{'kind': 'ByteArray',
                          'type': ihm.format_bcif._Uint8}],
        'stringData': string_data,
        'offsetEncoding': [{'kind': 'ByteArray',
                            'type': ihm.format_bcif._Uint8}],
        'offsets': offsets}
    d = {'data': indices,
         'encoding': [string_array_encoding]}
    return d, mask


class Category:
    def __init__(self, name, data):
        self.name = name
        self.data = data

    def _encode_rows(self, rows):
        if len(rows) and isinstance(rows[0], int):
            return _encode_int(rows)
        elif len(rows) and isinstance(rows[0], float):
            return _encode_float(rows)
        else:
            return _encode(rows)

    def get_bcif(self):
        nrows = 0
        cols = []
        for name, rows in self.data.items():
            nrows = len(rows)
            data, mask = self._encode_rows(rows)
            cols.append({'mask': mask, 'name': name,
                         'data': data})
        return {'name': self.name,
                'columns': cols, 'rowCount': nrows}


class Block(list):
    pass


class _BadMsgPackType:
    pass


class _RawMsgPack:
    def __init__(self, val):
        self._val = val


BAD_MSGPACK_TYPE = _BadMsgPackType()


def _add_msgpack(d, fh):
    """Add `d` to filelike object `fh` in msgpack format"""
    if isinstance(d, dict):
        fh.write(struct.pack('>Bi', 0xdf, len(d)))
        for key, val in d.items():
            _add_msgpack(key, fh)
            _add_msgpack(val, fh)
    elif isinstance(d, list):
        fh.write(struct.pack('>Bi', 0xdd, len(d)))
        for val in d:
            _add_msgpack(val, fh)
    elif isinstance(d, str):
        b = d.encode('utf8')
        fh.write(struct.pack('>Bi', 0xdb, len(b)))
        fh.write(b)
    elif isinstance(d, bytes):
        fh.write(struct.pack('>Bi', 0xc6, len(d)))
        fh.write(d)
    elif isinstance(d, int):
        fh.write(struct.pack('>Bi', 0xce, d))
    elif isinstance(d, float):
        fh.write(struct.pack('>Bf', 0xca, d))
    elif isinstance(d, _RawMsgPack):
        fh.write(d._val)
    elif d is None:
        fh.write(b'\xc0')
    elif d is BAD_MSGPACK_TYPE:
        # 0xc1 is not used in msgpack
        fh.write(b'\xc1')
    else:
        raise TypeError("Cannot handle %s" % type(d))


def _make_bcif_file(blocks):
    blocks = [{'header': 'ihm',
               'categories': [c.get_bcif() for c in block]}
              for block in blocks]
    d = {'version': '0.1', 'encoder': 'python-ihm test suite',
         'dataBlocks': blocks}
    return _python_to_msgpack(d)


def _python_to_msgpack(d):
    if _format:
        # Convert Python object `d` into msgpack format for the C-accelerated
        # parser
        fh = BytesIO()
        _add_msgpack(d, fh)
        fh.seek(0)
        return fh
    else:
        # Pure Python reader uses mocked-out msgpack to work on `d` directly
        return d


class Tests(unittest.TestCase):
    def test_decoder_base(self):
        """Test Decoder base class"""
        d = ihm.format_bcif._Decoder()
        self.assertIsNone(d._kind)
        d(enc=None, data=None)  # noop

    def test_string_array_decoder(self):
        """Test StringArray decoder"""
        d = ihm.format_bcif._StringArrayDecoder()
        self.assertEqual(d._kind, 'StringArray')

        # Int8 is signed char (so FF is -1)
        enc = {'stringData': 'aAB',
               'dataEncoding': [{'kind': 'ByteArray',
                                 'type': ihm.format_bcif._Int8}],
               'offsetEncoding': [{'kind': 'ByteArray',
                                   'type': ihm.format_bcif._Int8}],
               'offsets': b'\x00\x01\x03'}
        data = b'\x00\x01\x00\xFF'

        data = d(enc, data)
        self.assertEqual(list(data), ['a', 'AB', 'a', None])

    def test_byte_array_decoder(self):
        """Test ByteArray decoder"""
        d = ihm.format_bcif._ByteArrayDecoder()
        self.assertEqual(d._kind, 'ByteArray')

        # type 1 (signed char)
        data = d({'type': ihm.format_bcif._Int8}, b'\x00\x01\xFF')
        self.assertEqual(list(data), [0, 1, -1])

        # type 2 (signed short)
        data = d({'type': ihm.format_bcif._Int16}, b'\x00\x01\x01\xAC')
        self.assertEqual(list(data), [256, -21503])

        # type 3 (signed int)
        data = d({'type': ihm.format_bcif._Int32}, b'\x00\x01\x01\x05')
        self.assertEqual(list(data), [83951872])

        # type 4 (unsigned char)
        data = d({'type': ihm.format_bcif._Uint8}, b'\x00\xFF')
        self.assertEqual(list(data), [0, 255])

        # type 5 (unsigned short)
        data = d({'type': ihm.format_bcif._Uint16}, b'\x00\x01\x01\xAC')
        self.assertEqual(list(data), [256, 44033])

        # type 6 (unsigned int)
        data = d({'type': ihm.format_bcif._Uint32}, b'\x00\x01\x01\xFF')
        self.assertEqual(list(data), [4278255872])

        # type 32 (32-bit float)
        data = d({'type': ihm.format_bcif._Float32}, b'\x00\x00(B')
        self.assertAlmostEqual(list(data)[0], 42.0, delta=0.1)

        # type 33 (64-bit float)
        data = d({'type': ihm.format_bcif._Float64},
                 b'\x00\x00\x00\x00\x00\x00E@')
        self.assertAlmostEqual(list(data)[0], 42.0, delta=0.1)

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_byte_array_decoder_full_file(self):
        """Test ByteArray decoder working on full BinaryCIF file"""
        class MyCategory(Category):
            def __init__(self, name, data, raw_data, data_type, enc=None):
                Category.__init__(self, name, data)
                self.raw_data, self.data_type = raw_data, data_type
                self.enc = enc

            def _encode_rows(self, rows):
                if self.enc is not None:
                    enc = self.enc
                else:
                    enc = [{'kind': 'ByteArray', 'type': self.data_type}]
                return {'data': self.raw_data,
                        'encoding': enc}, None

        def get_decoded(data_type, raw_data, enc=None):
            cat = MyCategory('_exptl', {'method': []}, raw_data, data_type,
                             enc=enc)
            h = GenericHandler()
            self._read_bcif([Block([cat])], {'_exptl': h})
            return [x['method'] for x in h.data]

        # type 3 (signed int)
        data = get_decoded(ihm.format_bcif._Int32, b'\x00\x01\x01\x05')
        self.assertEqual(data, ['83951872'])

        # Raw data not a multiple of type size
        self.assertRaises(_format.FileFormatError, get_decoded,
                          ihm.format_bcif._Int16, b'\x00\x01\x01')
        self.assertRaises(_format.FileFormatError, get_decoded,
                          ihm.format_bcif._Uint16, b'\x00\x01\x01')
        self.assertRaises(_format.FileFormatError, get_decoded,
                          ihm.format_bcif._Int32, b'\x00\x01\x01')
        self.assertRaises(_format.FileFormatError, get_decoded,
                          ihm.format_bcif._Uint32, b'\x00\x01\x01')
        self.assertRaises(_format.FileFormatError, get_decoded,
                          ihm.format_bcif._Float64, b'\x00\x00\x00\x00')
        self.assertRaises(_format.FileFormatError, get_decoded,
                          ihm.format_bcif._Float32, b'\x00\x00\x00')

        # ByteArray must take raw data, not the output of another decoder
        self.assertRaises(_format.FileFormatError, get_decoded,
                          ihm.format_bcif._Int32, b'\x00\x01\x01\x05',
                          enc=[{'kind': 'ByteArray',
                                'type': ihm.format_bcif._Int32}] * 2)

        # type 1 (signed char)
        data = get_decoded(ihm.format_bcif._Int8, struct.pack('2b', -10, 100))
        self.assertEqual(data, ['-10', '100'])

        # type 4 (unsigned char)
        data = get_decoded(ihm.format_bcif._Uint8, b'\x00\xFF')
        self.assertEqual(data, ['0', '255'])

        # type 2 (16-bit signed int)
        data = get_decoded(ihm.format_bcif._Int16, struct.pack('<2h', -5, 10))
        self.assertEqual(data, ['-5', '10'])

        # type 5 (16-bit unsigned int)
        data = get_decoded(ihm.format_bcif._Uint16, struct.pack('<H', 50))
        self.assertEqual(data, ['50'])

        # type 33 (64-bit float)
        data = get_decoded(ihm.format_bcif._Float64,
                           b'\x00\x00\x00\x00\x00\x00E@')
        self.assertIsInstance(data[0], str)
        self.assertAlmostEqual(float(data[0]), 42.0, delta=0.1)

        # type 32 (32-bit float)
        data = get_decoded(ihm.format_bcif._Float32, b'\x00\x00(B')
        self.assertIsInstance(data[0], str)
        self.assertAlmostEqual(float(data[0]), 42.0, delta=0.1)

        # unsupported type (32-bit unsigned int)
        self.assertRaises(_format.FileFormatError,
                          get_decoded, ihm.format_bcif._Uint32,
                          struct.pack('<I', 5))

    def test_integer_packing_decoder_signed(self):
        """Test IntegerPacking decoder with signed data"""
        d = ihm.format_bcif._IntegerPackingDecoder()
        self.assertEqual(d._kind, 'IntegerPacking')

        # 1-byte data
        data = d({'isUnsigned': False, 'byteCount': 1},
                 [1, 2, -3, 127, 1, -128, -5])
        self.assertEqual(list(data), [1, 2, -3, 128, -133])

        # 2-byte data
        data = d({'isUnsigned': False, 'byteCount': 2},
                 [1, 2, -3, 32767, 1, -32768, -5])
        self.assertEqual(list(data), [1, 2, -3, 32768, -32773])

    def test_integer_packing_decoder_unsigned(self):
        """Test IntegerPacking decoder with unsigned data"""
        d = ihm.format_bcif._IntegerPackingDecoder()
        self.assertEqual(d._kind, 'IntegerPacking')

        # 1-byte data
        data = d({'isUnsigned': True, 'byteCount': 1},
                 [1, 2, 3, 127, 1, 255, 1])
        self.assertEqual(list(data), [1, 2, 3, 127, 1, 256])

        # 2-byte data
        data = d({'isUnsigned': True, 'byteCount': 2},
                 [1, 2, 3, 32767, 1, 65535, 5])
        self.assertEqual(list(data), [1, 2, 3, 32767, 1, 65540])

    def test_delta_decoder(self):
        """Test Delta decoder"""
        d = ihm.format_bcif._DeltaDecoder()
        self.assertEqual(d._kind, 'Delta')

        data = d({'origin': 1000}, [0, 3, 2, 1])
        self.assertEqual(list(data), [1000, 1003, 1005, 1006])

    def test_run_length_decoder(self):
        """Test RunLength decoder"""
        d = ihm.format_bcif._RunLengthDecoder()
        self.assertEqual(d._kind, 'RunLength')

        data = d({}, [1, 3, 2, 1, 3, 2])
        self.assertEqual(list(data), [1, 1, 1, 2, 3, 3])

    def test_fixed_point_decoder(self):
        """Test FixedPoint decoder"""
        d = ihm.format_bcif._FixedPointDecoder()
        self.assertEqual(d._kind, 'FixedPoint')

        data = list(d({'factor': 100}, [120, 123, 12]))
        self.assertEqual(len(data), 3)
        self.assertAlmostEqual(data[0], 1.20, delta=0.01)
        self.assertAlmostEqual(data[1], 1.23, delta=0.01)
        self.assertAlmostEqual(data[2], 0.12, delta=0.01)

    def test_interval_quantization_decoder(self):
        """Test IntervalQuantization decoder"""
        d = ihm.format_bcif._IntervalQuantizationDecoder()
        self.assertEqual(d._kind, 'IntervalQuantization')

        data = list(d({'min': 1.0, 'max': 2.0, 'numSteps': 3},
                      [0, 1, 2]))
        self.assertEqual(len(data), 3)
        self.assertAlmostEqual(data[0], 1.0, delta=0.01)
        self.assertAlmostEqual(data[1], 1.5, delta=0.01)
        self.assertAlmostEqual(data[2], 2.0, delta=0.01)

    def test_decode(self):
        """Test _decode function"""
        data = b'\x01\x03\x02\x01\x03\x02'
        runlen = {'kind': 'RunLength'}
        bytearr = {'kind': 'ByteArray', 'type': ihm.format_bcif._Int8}
        data = ihm.format_bcif._decode(data, [runlen, bytearr])
        self.assertEqual(list(data), [1, 1, 1, 2, 3, 3])

    def _read_bcif(self, blocks, category_handlers,
                   unknown_category_handler=None,
                   unknown_keyword_handler=None):
        fh = _make_bcif_file(blocks)
        sys.modules['msgpack'] = MockMsgPack
        r = ihm.format_bcif.BinaryCifReader(fh, category_handlers,
                                            unknown_category_handler,
                                            unknown_keyword_handler)
        r.read_file()

    def test_category_case_insensitive(self):
        """Categories and keywords should be case insensitive"""
        cat1 = Category('_exptl', {'method': ['foo']})
        cat2 = Category('_Exptl', {'METHod': ['foo']})
        for cat in cat1, cat2:
            h = GenericHandler()
            self._read_bcif([Block([cat])], {'_exptl': h})
        self.assertEqual(h.data, [{'method': 'foo'}])

    def test_int_keys(self):
        """Check handling of integer keywords"""
        cat = Category('_foo', {'intkey1': [42]})
        h = GenericHandler()
        self._read_bcif([Block([cat])], {'_foo': h})
        self.assertEqual(h.data, [{'intkey1': 42}])

        # Float will be truncated to int
        cat = Category('_foo', {'intkey1': [42.34]})
        h = GenericHandler()
        self._read_bcif([Block([cat])], {'_foo': h})
        self.assertEqual(h.data, [{'intkey1': 42}])

        # Can coerce an intlike string to int
        cat = Category('_foo', {'intkey1': ["42"]})
        h = GenericHandler()
        self._read_bcif([Block([cat])], {'_foo': h})
        self.assertEqual(h.data, [{'intkey1': 42}])

        # Cannot coerce arbitrary string to int
        cat = Category('_foo', {'intkey1': ["some string"]})
        h = GenericHandler()
        self.assertRaises(ValueError, self._read_bcif,
                          [Block([cat])], {'_foo': h})

    def test_float_keys(self):
        """Check handling of floating-point keywords"""
        cat = Category('_foo', {'floatkey1': [42.340]})
        h = GenericHandler()
        self._read_bcif([Block([cat])], {'_foo': h})
        val = h.data[0]['floatkey1']
        self.assertIsInstance(val, float)
        self.assertAlmostEqual(val, 42.34, delta=0.01)

        # int will be coerced to float
        cat = Category('_foo', {'floatkey1': [42]})
        h = GenericHandler()
        self._read_bcif([Block([cat])], {'_foo': h})
        val = h.data[0]['floatkey1']
        self.assertIsInstance(val, float)
        self.assertAlmostEqual(val, 42.0, delta=0.01)

        # Can coerce a floatlike string to float
        cat = Category('_foo', {'floatkey1': ["42.340"]})
        h = GenericHandler()
        self._read_bcif([Block([cat])], {'_foo': h})
        val = h.data[0]['floatkey1']
        self.assertIsInstance(val, float)
        self.assertAlmostEqual(val, 42.34, delta=0.01)

        # Cannot coerce arbitrary string to float
        cat = Category('_foo', {'floatkey1': ["some string"]})
        h = GenericHandler()
        self.assertRaises(ValueError, self._read_bcif,
                          [Block([cat])], {'_foo': h})

    def test_bool_keys(self):
        """Check handling of boolean keywords"""
        cat = Category('_foo', {'boolkey1': ["YES", "NO", "GARBAGE"]})
        h = GenericHandler()
        self._read_bcif([Block([cat])], {'_foo': h})
        self.assertEqual(h.data, [{'boolkey1': True},
                                  {'boolkey1': False},
                                  {}])
        h = GenericHandler()
        h.omitted = 'OM'
        self._read_bcif([Block([cat])], {'_foo': h})
        self.assertEqual(h.data, [{'boolkey1': True},
                                  {'boolkey1': False},
                                  {'boolkey1': 'OM'}])

        # Float will be mapped to omitted (None, or handler.omitted)
        cat = Category('_foo', {'boolkey1': [42.34]})
        h = GenericHandler()
        self._read_bcif([Block([cat])], {'_foo': h})
        self.assertEqual(h.data, [{}])
        h = GenericHandler()
        h.omitted = 'OM'
        self._read_bcif([Block([cat])], {'_foo': h})
        self.assertEqual(h.data, [{'boolkey1': 'OM'}])

        # Int will be mapped to omitted (None)
        cat = Category('_foo', {'boolkey1': [42]})
        h = GenericHandler()
        self._read_bcif([Block([cat])], {'_foo': h})
        self.assertEqual(h.data, [{}])
        h = GenericHandler()
        h.omitted = 'OM'
        self._read_bcif([Block([cat])], {'_foo': h})
        self.assertEqual(h.data, [{'boolkey1': 'OM'}])

    def test_omitted_unknown(self):
        """Test handling of omitted/unknown data"""
        cat = Category('_foo',
                       {'var1': ['test1', '?', 'test2', None, 'test3']})
        h = GenericHandler()
        self._read_bcif([Block([cat])], {'_foo': h})
        self.assertEqual(h.data,
                         [{'var1': 'test1'}, {'var1': '?'},
                          {'var1': 'test2'}, {}, {'var1': 'test3'}])

    def _read_bcif_raw(self, d, category_handlers):
        fh = _python_to_msgpack(d)
        r = ihm.format_bcif.BinaryCifReader(fh, category_handlers)
        r.read_file()

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_read_header_c(self):
        """Test handling of various BinaryCIF headers"""
        # No header
        d = 42
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw, d, {})
        # Header keys not strings
        d = {42: 50}
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw, d, {})
        # Data blocks not a list
        d = {'dataBlocks': 42}
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw, d, {})
        # Empty header
        d = {}
        self._read_bcif_raw(d, {})
        # Unknown keys should be ignored
        d = {'unknown-key': None}
        self._read_bcif_raw(d, {})
        # Skipped object of bad type
        d = {'unknown-keyword': BAD_MSGPACK_TYPE}
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw, d, {})

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_bad_block(self):
        """Test handling of various bad BinaryCIF blocks"""
        # Block not a map
        d = {'dataBlocks': [42]}
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw, d, {})
        # Block keys not strings
        d = {'dataBlocks': [{42: 50}]}
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw, d, {})
        # Skipped object of bad type
        d = {'dataBlocks': [{'unknown-keyword': BAD_MSGPACK_TYPE}]}
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw, d, {})

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_bad_categories(self):
        """Test handling of various bad BinaryCIF category lists"""
        # Categories not a list
        d = {'dataBlocks': [{'categories': 42}]}
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw, d, {})

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_bad_category(self):
        """Test handling of various bad BinaryCIF categories"""
        def make_bcif(c):
            return {'dataBlocks': [{'categories': [c]}]}

        # Category not a map
        d = make_bcif(42)
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw, d, {})
        # Category keys not strings
        d = make_bcif({42: 50})
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw, d, {})
        # Category name not a string
        d = make_bcif({'name': 42})
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw, d, {})
        # Skipped object of bad type
        d = make_bcif({'name': 'bar', 'unknown-key': BAD_MSGPACK_TYPE})
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw, d, {})

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_bad_columns(self):
        """Test handling of various bad BinaryCIF column lists"""
        def make_bcif(c):
            return {'dataBlocks': [{'categories': [{'name': '_foo',
                                                    'columns': c}]}]}

        # Columns are skipped
        d = make_bcif([])
        self._read_bcif_raw(d, {})

        # Columns not an array
        d = make_bcif(42)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_read_column(self):
        """Test handling of various BinaryCIF columns"""
        def make_bcif(c):
            return {'dataBlocks': [{'categories': [{'name': '_foo',
                                                    'columns': [c]}]}]}

        # Column not a map
        d = make_bcif(42)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})
        # Column keys not strings
        d = make_bcif({42: 50})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})
        # Column name not string
        d = make_bcif({'name': 50})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Skipped unknown key contains invalid data
        d = make_bcif({'name': 'bar',
                       'data': {'data': b'\x00\x00\x00\x00',
                                'encoding':
                                [{'kind': 'ByteArray',
                                  'type': ihm.format_bcif._Int32}]},
                       'unknown-key': BAD_MSGPACK_TYPE})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Normal usage; unknown keys should be skipped
        d = make_bcif({'name': 'bar',
                       'data': {'data': b'\x00\x00\x00\x00',
                                'encoding':
                                [{'kind': 'ByteArray',
                                  'type': ihm.format_bcif._Int32}]},
                       'unknown-key': None})
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_read_mask(self):
        """Test handling of BinaryCIF mask"""
        def make_bcif(mask):
            c = {'name': 'bar',
                 'data': {'data': b'\x00\x00\x00\x00',
                          'encoding':
                          [{'kind': 'ByteArray',
                            'type': ihm.format_bcif._Int32}]},
                 'mask': mask}
            return {'dataBlocks': [{'categories': [{'name': '_foo',
                                                    'columns': [c]}]}]}
        # Bad mask type
        d = make_bcif('foo')
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Null mask
        d = make_bcif(None)
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})

        # Error trying to read mask map (or nil)
        d = make_bcif(BAD_MSGPACK_TYPE)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Mask of bad type
        d = make_bcif("foo")
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Map keys not strings
        d = make_bcif({42: 50})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Encodings not an array
        d = make_bcif({'encoding': None})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Bad data
        d = make_bcif({'data': None})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Skipped unknown key contains invalid data
        d = make_bcif({'unknown-key': BAD_MSGPACK_TYPE})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Unknown map keys should be ignored
        d = make_bcif({'unknown-key': None})
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})

        # Bad data type
        d = make_bcif({'data': b'\x00\x00\x01\x00\x02\x00',
                       'encoding': [{'kind': 'ByteArray',
                                     'type': ihm.format_bcif._Uint16}]})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Normal usage
        d = make_bcif({'data': b'\x00\x01\x02',
                       'encoding': [{'kind': 'ByteArray',
                                     'type': ihm.format_bcif._Uint8}]})
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_read_data(self):
        """Test handling of BinaryCIF data"""
        def make_bcif(data):
            c = {'name': 'bar',
                 'data': data}
            return {'dataBlocks': [{'categories': [{'name': '_foo',
                                                    'columns': [c]}]}]}
        # Data not a map
        d = make_bcif('foo')
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Map keys not strings
        d = make_bcif({42: 50})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Bad data
        d = make_bcif({'data': None})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Skipped unknown key contains invalid data
        d = make_bcif({'unknown-key': BAD_MSGPACK_TYPE})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Unknown map keys should be ignored
        d = make_bcif({'unknown-key': None,
                       'data': b'\x00\x00\x00\x00',
                       'encoding': [{'kind': 'ByteArray',
                                     'type': ihm.format_bcif._Int32}]})
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_read_encoding(self):
        """Test handling of BinaryCIF encoding information"""
        def make_bcif(encoding):
            c = {'name': 'bar',
                 'data': {'data': b'\x00\x00\x00\x00',
                          'encoding': [encoding]}}
            return {'dataBlocks': [{'categories': [{'name': '_foo',
                                                    'columns': [c]}]}]}
        # Encoding not a map
        d = make_bcif('foo')
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Map keys not strings
        d = make_bcif({42: 50})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Bad kind
        d = make_bcif({'kind': None})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Bad dataEncoding
        d = make_bcif({'dataEncoding': None})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Bad offsetEncoding
        d = make_bcif({'offsetEncoding': None})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Bad stringData
        d = make_bcif({'stringData': None})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Bad offsets
        d = make_bcif({'offsets': None})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Bad type
        d = make_bcif({'type': None})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Skipped unknown key contains invalid data
        d = make_bcif({'unknown-key': BAD_MSGPACK_TYPE})
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Normal usage; unknown keys should be ignored
        d = make_bcif({'kind': 'ByteArray',
                       'type': ihm.format_bcif._Int32,
                       'unknown-key': None})
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_string_array_encoding_c(self):
        """Test handling of various BinaryCIF StringArray encodings"""
        def make_bcif(data, data_type, offsets, offsets_type,
                      data_encoding='ByteArray'):
            c = {'name': 'bar',
                 'data': {'data': data,
                          'encoding':
                          [{'kind': 'StringArray', 'stringData': 'aAB',
                            'dataEncoding': [{'kind': data_encoding,
                                              'type': data_type}],
                            'offsetEncoding': [{'kind': 'ByteArray',
                                                'type': offsets_type}],
                            'offsets': offsets}]}}
            return {'dataBlocks': [{'categories': [{'name': '_foo',
                                                    'columns': [c]}]}]}

        # Test normal usage
        for (data, data_type) in (
                (b'\x00\x01', ihm.format_bcif._Uint8),
                (b'\x00\x01', ihm.format_bcif._Int8),
                (b'\x00\x00\x01\x00', ihm.format_bcif._Uint16),
                (b'\x00\x00\x01\x00', ihm.format_bcif._Int16),
                (b'\x00\x00\x00\x00\x01\x00\x00\x00', ihm.format_bcif._Int32)):
            d = make_bcif(data=data, data_type=data_type,
                          offsets=b'\x00\x01\x03',
                          offsets_type=ihm.format_bcif._Uint8)
            h = GenericHandler()
            self._read_bcif_raw(d, {'_foo': h})
            self.assertEqual(h.data, [{'bar': 'a'}, {'bar': 'AB'}])

        # StringArray decoding can't be used for offset or data
        d = make_bcif(data=b'\x00\x01', data_type=ihm.format_bcif._Uint8,
                      offsets=b'\x00\x01\x03',
                      offsets_type=ihm.format_bcif._Uint8,
                      data_encoding='StringArray')
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Indices must be int, not float
        d = make_bcif(data=b'\x00\x00(B', data_type=ihm.format_bcif._Float32,
                      offsets=b'\x00\x01\x03',
                      offsets_type=ihm.format_bcif._Uint8)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Offsets must be int, not float
        d = make_bcif(data=b'\x00\x01', data_type=ihm.format_bcif._Uint8,
                      offsets=b'\x00\x00(B',
                      offsets_type=ihm.format_bcif._Float32)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Offsets must be in range of string data
        d = make_bcif(data=b'\x00\x01', data_type=ihm.format_bcif._Uint8,
                      offsets=b'\x00\x01\xcc',
                      offsets_type=ihm.format_bcif._Uint8)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Out-of-range indices return empty strings
        for data in (struct.pack('b', 40), struct.pack('b', -32)):
            d = make_bcif(data=data, data_type=ihm.format_bcif._Int8,
                          offsets=b'\x00\x01\x03',
                          offsets_type=ihm.format_bcif._Uint8)
            h = GenericHandler()
            self._read_bcif_raw(d, {'_foo': h})
            self.assertEqual(h.data, [{'bar': ''}])

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_fixed_point_encoding_c(self):
        """Test handling of various BinaryCIF FixedPoint encodings"""
        def make_bcif(data, data_type, factor):
            c = {'name': 'bar',
                 'data': {'data': data,
                          'encoding':
                          [{'kind': 'FixedPoint', 'factor': factor},
                           {'kind': 'ByteArray', 'type': data_type}]}}
            return {'dataBlocks': [{'categories': [{'name': '_foo',
                                                    'columns': [c]}]}]}

        # Test normal usage
        for (data, data_type) in (
                (struct.pack('b', 104), ihm.format_bcif._Int8),
                (struct.pack('B', 104), ihm.format_bcif._Uint8),
                (struct.pack('<h', 104), ihm.format_bcif._Int16),
                (struct.pack('<H', 104), ihm.format_bcif._Uint16),
                (struct.pack('<i', 104), ihm.format_bcif._Int32),
                (struct.pack('<I', 104), ihm.format_bcif._Uint32)):
            d = make_bcif(data=data, data_type=data_type, factor=100)
            h = GenericHandler()
            self._read_bcif_raw(d, {'_foo': h})
            bar = h.data[0]['bar']
            self.assertIsInstance(bar, str)
            self.assertAlmostEqual(float(bar), 1.04, delta=0.01)

        # Bad factor type
        d = make_bcif(data=struct.pack('<i', 204),
                      data_type=ihm.format_bcif._Int32,
                      factor='bad factor')
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Bad input type
        d = make_bcif(data=struct.pack('<f', 42.0),
                      data_type=ihm.format_bcif._Float32, factor=100)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_run_length_encoding_c(self):
        """Test handling of various BinaryCIF RunLength encodings"""
        def make_bcif(data, data_type):
            c = {'name': 'bar',
                 'data': {'data': data,
                          'encoding':
                          [{'kind': 'RunLength'},
                           {'kind': 'ByteArray', 'type': data_type}]}}
            return {'dataBlocks': [{'categories': [{'name': '_foo',
                                                    'columns': [c]}]}]}

        # Test normal usage with signed 32-bit integer data
        d = make_bcif(data=struct.pack('<2i', 5, 3),
                      data_type=ihm.format_bcif._Int32)
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})
        self.assertEqual(h.data, [{'bar': '5'}] * 3)

        # Test normal usage with signed 8-bit integer data
        d = make_bcif(data=struct.pack('2b', 5, 3),
                      data_type=ihm.format_bcif._Int8)
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})
        self.assertEqual(h.data, [{'bar': '5'}] * 3)

        # Test normal usage with unsigned 8-bit integer data
        d = make_bcif(data=struct.pack('2B', 5, 3),
                      data_type=ihm.format_bcif._Uint8)
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})
        self.assertEqual(h.data, [{'bar': '5'}] * 3)

        # Test normal usage with signed 16-bit integer data
        d = make_bcif(data=struct.pack('<2h', 5, 3),
                      data_type=ihm.format_bcif._Int16)
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})
        self.assertEqual(h.data, [{'bar': '5'}] * 3)

        # Test normal usage with unsigned 16-bit integer data
        d = make_bcif(data=struct.pack('<2H', 5, 3),
                      data_type=ihm.format_bcif._Uint16)
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})
        self.assertEqual(h.data, [{'bar': '5'}] * 3)

        # Data size should be even
        d = make_bcif(data=struct.pack('<3H', 5, 3, 8),
                      data_type=ihm.format_bcif._Uint16)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Bad input type
        d = make_bcif(data=struct.pack('<2f', 42.0, 1.0),
                      data_type=ihm.format_bcif._Float32)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Negative counts should be rejected
        d = make_bcif(data=struct.pack('<2i', 5, -3),
                      data_type=ihm.format_bcif._Int32)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Very large positive counts should be rejected
        d = make_bcif(data=struct.pack('<2i', 5, int(1e8)),
                      data_type=ihm.format_bcif._Int32)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_delta_encoding_c(self):
        """Test handling of various BinaryCIF Delta encodings"""
        def make_bcif(data, data_type, origin):
            c = {'name': 'bar',
                 'data': {'data': data,
                          'encoding':
                          [{'kind': 'Delta', 'origin': origin},
                           {'kind': 'ByteArray', 'type': data_type}]}}
            return {'dataBlocks': [{'categories': [{'name': '_foo',
                                                    'columns': [c]}]}]}

        # Test normal usage, 32-bit signed int
        d = make_bcif(data=struct.pack('<2i', 5, -2),
                      data_type=ihm.format_bcif._Int32, origin=50)
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})
        self.assertEqual(h.data, [{'bar': '55'}, {'bar': '53'}])

        # Test normal usage, 8-bit signed int
        d = make_bcif(data=struct.pack('2b', 5, -2),
                      data_type=ihm.format_bcif._Int8, origin=50)
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})
        self.assertEqual(h.data, [{'bar': '55'}, {'bar': '53'}])

        # Test normal usage, 8-bit unsigned int
        d = make_bcif(data=struct.pack('2B', 5, 20),
                      data_type=ihm.format_bcif._Uint8, origin=50)
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})
        self.assertEqual(h.data, [{'bar': '55'}, {'bar': '75'}])

        # Test normal usage, 16-bit signed int
        d = make_bcif(data=struct.pack('<2h', 5, -2),
                      data_type=ihm.format_bcif._Int16, origin=50)
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})
        self.assertEqual(h.data, [{'bar': '55'}, {'bar': '53'}])

        # Test normal usage, 16-bit unsigned int
        d = make_bcif(data=struct.pack('<2H', 5, 20),
                      data_type=ihm.format_bcif._Uint16, origin=50)
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})
        self.assertEqual(h.data, [{'bar': '55'}, {'bar': '75'}])

        # Bad input type
        d = make_bcif(data=struct.pack('<f', 42.0),
                      data_type=ihm.format_bcif._Float32, origin=50)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Bad origin type
        d = make_bcif(data=struct.pack('<2i', 5, 3),
                      data_type=ihm.format_bcif._Int32, origin='foo')
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_integer_packing_encoding_c(self):
        """Test handling of various BinaryCIF IntegerPacking encodings"""
        def make_bcif(data, data_type):
            c = {'name': 'bar',
                 'data': {'data': data,
                          'encoding':
                          [{'kind': 'IntegerPacking'},
                           {'kind': 'ByteArray', 'type': data_type}]}}
            return {'dataBlocks': [{'categories': [{'name': '_foo',
                                                    'columns': [c]}]}]}

        # Test signed 8-bit input
        d = make_bcif(data=struct.pack('6b', 5, 127, 8, -30, -128, -10),
                      data_type=ihm.format_bcif._Int8)
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})
        self.assertEqual(h.data, [{'bar': '5'}, {'bar': '135'},
                                  {'bar': '-30'}, {'bar': '-138'}])

        # Test unsigned 8-bit input
        d = make_bcif(data=struct.pack('3B', 5, 255, 8),
                      data_type=ihm.format_bcif._Uint8)
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})
        self.assertEqual(h.data, [{'bar': '5'}, {'bar': '263'}])

        # Test signed 16-bit input
        d = make_bcif(data=struct.pack('<6h', 5, 32767, 8, -30, -32768, -10),
                      data_type=ihm.format_bcif._Int16)
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})
        self.assertEqual(h.data, [{'bar': '5'}, {'bar': '32775'},
                                  {'bar': '-30'}, {'bar': '-32778'}])

        # Test unsigned 16-bit input
        d = make_bcif(data=struct.pack('<3H', 5, 65535, 8),
                      data_type=ihm.format_bcif._Uint16)
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})
        self.assertEqual(h.data, [{'bar': '5'}, {'bar': '65543'}])

        # Bad input type
        d = make_bcif(data=struct.pack('<f', 42.0),
                      data_type=ihm.format_bcif._Float32)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_interval_quantization_encoding_c(self):
        """Test handling of various BinaryCIF IntervalQuantization encodings"""
        def make_bcif(data, data_type, minval=1, maxval=2, numsteps=11):
            c = {'name': 'bar',
                 'data': {'data': data,
                          'encoding':
                          [{'kind': 'IntervalQuantization',
                            'min': minval, 'max': maxval,
                            'numSteps': numsteps},
                           {'kind': 'ByteArray', 'type': data_type}]}}
            return {'dataBlocks': [{'categories': [{'name': '_foo',
                                                    'columns': [c]}]}]}

        # Test normal usage
        for (data, data_type) in (
                (struct.pack('3b', 0, 1, 2), ihm.format_bcif._Int8),
                (struct.pack('3B', 0, 1, 2), ihm.format_bcif._Uint8),
                (struct.pack('<3h', 0, 1, 2), ihm.format_bcif._Int16),
                (struct.pack('<3H', 0, 1, 2), ihm.format_bcif._Uint16),
                (struct.pack('<3i', 0, 1, 2), ihm.format_bcif._Int32),
                (struct.pack('<3I', 0, 1, 2), ihm.format_bcif._Uint32)):
            d = make_bcif(data=data, data_type=data_type,
                          minval=1.0, maxval=2.0, numsteps=3)
            h = GenericHandler()
            self._read_bcif_raw(d, {'_foo': h})
            for got, want in zip((x['bar'] for x in h.data), [1.0, 1.5, 2.0]):
                self.assertIsInstance(got, str)
                self.assertAlmostEqual(float(got), want, delta=0.01)

        # Min, max int (not float)
        d = make_bcif(data=struct.pack('3b', 0, 1, 2),
                      data_type=ihm.format_bcif._Int8,
                      minval=1, maxval=2, numsteps=3)
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})
        for got, want in zip((x['bar'] for x in h.data), [1.0, 1.5, 2.0]):
            self.assertIsInstance(got, str)
            self.assertAlmostEqual(float(got), want, delta=0.01)

        # min can be any msgpack int or float type
        for raw, positive in (
                (struct.pack('B', 100), True),             # positive fixint
                (struct.pack('B', 0xe0 | 22), False),      # negative fixint
                (struct.pack('2B', 0xcc, 100), True),      # uint8
                (struct.pack('>BH', 0xcd, 100), True),     # uint16
                (struct.pack('>BI', 0xce, 100), True),     # uint32
                (struct.pack('Bb', 0xd0, 100), True),      # int8
                (struct.pack('Bb', 0xd0, -10), False),
                (struct.pack('>Bh', 0xd1, 100), True),     # int16
                (struct.pack('>Bh', 0xd1, -10), False),
                (struct.pack('>Bi', 0xd2, 100), True),     # int32
                (struct.pack('>Bi', 0xd2, -10), False),
                (struct.pack('>Bf', 0xca, 100.0), True),   # float32
                (struct.pack('>Bd', 0xcb, 100.0), True)):  # float64
            d = make_bcif(data=struct.pack('b', 0),
                          data_type=ihm.format_bcif._Int8,
                          minval=_RawMsgPack(raw), maxval=200, numsteps=3)
            h = GenericHandler()
            self._read_bcif_raw(d, {'_foo': h})
            bar = h.data[0]['bar']
            self.assertIsInstance(bar, str)
            self.assertAlmostEqual(float(bar), 100.0 if positive else -10.0,
                                   delta=0.01)

        # Bad number type
        d = make_bcif(data=struct.pack('b', 0),
                      data_type=ihm.format_bcif._Int8,
                      minval='foo', maxval=200, numsteps=3)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Bad input type
        d = make_bcif(data=struct.pack('<3f', 0.0, 1.0, 2.0),
                      data_type=ihm.format_bcif._Float32)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Error trying to read encoding min parameter
        d = make_bcif(data=struct.pack('3b', 0, 1, 2),
                      data_type=ihm.format_bcif._Int8,
                      minval=BAD_MSGPACK_TYPE)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Error trying to read encoding max parameter
        d = make_bcif(data=struct.pack('3b', 0, 1, 2),
                      data_type=ihm.format_bcif._Int8,
                      maxval=BAD_MSGPACK_TYPE)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Error trying to read encoding numsteps parameter
        d = make_bcif(data=struct.pack('3b', 0, 1, 2),
                      data_type=ihm.format_bcif._Int8,
                      numsteps='foo')
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

        # Bad number of steps
        d = make_bcif(data=struct.pack('3b', 0, 1, 2),
                      data_type=ihm.format_bcif._Int8,
                      numsteps=1)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_process_bcif_category_c(self):
        """Test processing of BinaryCIF category"""
        def make_bcif(data1, data2, data_type):
            c1 = {'name': 'bar',
                  'data': {'data': data1,
                           'encoding':
                           [{'kind': 'IntegerPacking'},
                            {'kind': 'ByteArray', 'type': data_type}]}}
            c2 = {'name': 'baz',
                  'data': {'data': data2,
                           'encoding':
                           [{'kind': 'IntegerPacking'},
                            {'kind': 'ByteArray', 'type': data_type}]}}
            return {'dataBlocks': [{'categories': [{'name': '_foo',
                                                    'columns': [c1, c2]}]}]}

        class _ThrowHandler(GenericHandler):
            def __call__(self, *args):
                raise ValueError("some error")

        # Normal operation
        d = make_bcif(data1=struct.pack('2b', 1, 42),
                      data2=struct.pack('2b', 8, 4),
                      data_type=ihm.format_bcif._Int8)
        h = GenericHandler()
        self._read_bcif_raw(d, {'_foo': h})
        self.assertEqual(h.data,
                         [{'bar': '1', 'baz': '8'}, {'bar': '42', 'baz': '4'}])

        # Handler errors should be propagated
        h = _ThrowHandler()
        self.assertRaises(ValueError, self._read_bcif_raw, d, {'_foo': h})

        # Mismatched column size
        d = make_bcif(data1=struct.pack('3b', 1, 42, 9),
                      data2=struct.pack('2b', 8, 4),
                      data_type=ihm.format_bcif._Int8)
        h = GenericHandler()
        self.assertRaises(_format.FileFormatError, self._read_bcif_raw,
                          d, {'_foo': h})

    @unittest.skipIf(_format is None, "No C tokenizer")
    def test_read_error(self):
        """Test handling of errors from filelike read()"""
        def make_bcif():
            c1 = {'name': 'bar',
                  'data': {'data': struct.pack('2b', 1, 42),
                           'encoding':
                           [{'kind': 'IntegerPacking'},
                            {'kind': 'ByteArray',
                             'type': ihm.format_bcif._Int8,
                             'min': 1.0}]},
                  'mask': None}
            return {'dataBlocks': [{'categories': [{'name': '_foo',
                                                    'columns': [c1]}]}]}

        class ReadError:
            """Filelike object that returns defined-size blocks from read,
               or errors out if empty"""
            def __init__(self, read_sz):
                self.read_sz = read_sz
                self.data = b''

            def write(self, b):
                self.data += b

            def read(self, sz):
                if not self.read_sz:
                    raise IndexError("foo")
                rsz = self.read_sz.pop(0)
                assert sz >= rsz
                ret = self.data[:rsz]
                self.data = self.data[rsz:]
                return ret

        def run_test(read_sz, ind=-1, cat='_foo'):
            d = make_bcif()
            h = GenericHandler()
            fh = ReadError(read_sz)
            _add_msgpack(d, fh)
            # If ind is given, just read up to the first instance of that
            # msgpack type
            if ind > 0:
                fh.read_sz = [fh.data.index(ind)]
            r = ihm.format_bcif.BinaryCifReader(fh, {cat: h})
            r.read_file()

        # Less data read than requested
        with self.assertRaises(OSError) as cm:
            run_test([0])
        self.assertIn('Less data read than requested', str(cm.exception))

        # Exception in read_map_or_nil (None=0xc0 in msgpack)
        self.assertRaises(IndexError, run_test, [], ind=0xc0)

        # Exception in read_array (array=0xdd)
        self.assertRaises(IndexError, run_test, [], ind=0xdd)

        # Exception in read_int (int=0xce)
        self.assertRaises(IndexError, run_test, [], ind=0xce)

        # Exception in read_bcif_exact_string (str=0xdb)
        self.assertRaises(IndexError, run_test, [], ind=0xdb)

        # Exception in read_bcif_any_double (float=0xca)
        self.assertRaises(IndexError, run_test, [], ind=0xca)

        # Exception in read_bcif_string_dup (107=index of column name's
        # string size)
        self.assertRaises(IndexError, run_test, [107])

        # Exception in read_bcif_string (168=index of encoding name's
        # string size)
        self.assertRaises(IndexError, run_test, [168])

        # Exception in read_map (25=index of second map)
        self.assertRaises(IndexError, run_test, [25])

        # Exception in read_binary (137=index of data's binary size)
        self.assertRaises(IndexError, run_test, [137])

        # Exception in skip_no_limit
        self.assertRaises(IndexError, run_test, [147], cat='_bar')

    def test_omitted_unknown_not_in_file_explicit(self):
        """Test explicit handling of omitted/unknown/not in file data"""
        cat = Category('_foo',
                       {'var1': ['test1', '?', 'test2', None, 'test3']})
        h = GenericHandler()
        h.omitted = 'OMIT'
        h.unknown = 'UNK'
        h.not_in_file = 'NOT'
        h._keys = ('var1', 'var2')
        h._int_keys = frozenset()
        h._float_keys = frozenset()
        h._bool_keys = frozenset()
        self._read_bcif([Block([cat])], {'_foo': h})
        self.assertEqual(h.data,
                         [{'var1': 'test1', 'var2': 'NOT'},
                          {'var1': 'UNK', 'var2': 'NOT'},
                          {'var1': 'test2', 'var2': 'NOT'},
                          {'var1': 'OMIT', 'var2': 'NOT'},
                          {'var1': 'test3', 'var2': 'NOT'}])

    def test_unknown_categories_ignored(self):
        """Check that unknown categories are just ignored"""
        cat1 = Category('_foo', {'var1': ['test1']})
        cat2 = Category('_bar', {'var2': ['test2']})
        h = GenericHandler()
        self._read_bcif([Block([cat1, cat2])], {'_foo': h})
        self.assertEqual(h.data, [{'var1': 'test1'}])

    def test_unknown_categories_handled(self):
        """Check that unknown categories are handled if requested"""
        class CatHandler:
            def __init__(self):
                self.warns = []

            def __call__(self, cat, line):
                self.warns.append((cat, line))

        ch = CatHandler()
        cat1 = Category('_foo', {'var1': ['test1']})
        cat2 = Category('_bar', {'var2': ['test2']})
        h = GenericHandler()
        self._read_bcif([Block([cat1, cat2])], {'_foo': h},
                        unknown_category_handler=ch)
        self.assertEqual(h.data, [{'var1': 'test1'}])
        self.assertEqual(ch.warns, [('_bar', 0)])

    def test_unknown_keywords_ignored(self):
        """Check that unknown keywords are ignored"""
        cat = Category('_foo', {'var1': ['test1'], 'othervar': ['test2']})
        h = GenericHandler()
        self._read_bcif([Block([cat])], {'_foo': h})
        self.assertEqual(h.data, [{'var1': 'test1'}])

    def test_unknown_keywords_handled(self):
        """Check that unknown keywords are handled if requested"""
        class KeyHandler:
            def __init__(self):
                self.warns = []

            def __call__(self, cat, key, line):
                self.warns.append((cat, key, line))

        kh = KeyHandler()
        cat = Category('_foo', {'var1': ['test1'], 'othervar': ['test2']})
        h = GenericHandler()
        self._read_bcif([Block([cat])], {'_foo': h},
                        unknown_keyword_handler=kh)
        self.assertEqual(h.data, [{'var1': 'test1'}])
        self.assertEqual(kh.warns, [('_foo', 'othervar', 0)])

    def test_multiple_data_blocks(self):
        """Test handling of multiple data blocks"""
        block1 = Block([Category('_foo',
                                 {'var1': ['test1'], 'var2': ['test2']})])
        block2 = Block([Category('_foo', {'var3': ['test3']})])
        fh = _make_bcif_file([block1, block2])

        h = GenericHandler()
        r = ihm.format_bcif.BinaryCifReader(fh, {'_foo': h})
        sys.modules['msgpack'] = MockMsgPack
        # Read first data block
        self.assertTrue(r.read_file())
        self.assertEqual(h.data, [{'var1': 'test1', 'var2': 'test2'}])

        # Read second data block
        h.data = []
        self.assertFalse(r.read_file())
        self.assertEqual(h.data, [{'var3': 'test3'}])

        # No more data blocks
        h.data = []
        self.assertFalse(r.read_file())
        self.assertEqual(h.data, [])

    def test_encoder(self):
        """Test _Encoder base class"""
        e = ihm.format_bcif._Encoder()
        e(None)  # noop

    def test_byte_array_encoder(self):
        """Test ByteArray encoder"""
        d = ihm.format_bcif._ByteArrayEncoder()

        # type 1 (signed char)
        data, encd = d([0, 1, -1])
        self.assertEqual(data, b'\x00\x01\xFF')
        self.assertEqual(encd, {'kind': 'ByteArray',
                                'type': ihm.format_bcif._Int8})

        # type 2 (signed short)
        data, encd = d([256, -21503])
        self.assertEqual(data, b'\x00\x01\x01\xAC')
        self.assertEqual(encd, {'kind': 'ByteArray',
                                'type': ihm.format_bcif._Int16})

        # type 3 (signed int)
        data, encd = d([-83951872])
        self.assertEqual(data, b'\x00\xff\xfe\xfa')
        self.assertEqual(encd, {'kind': 'ByteArray',
                                'type': ihm.format_bcif._Int32})

        # type 4 (unsigned char)
        data, encd = d([0, 255])
        self.assertEqual(data, b'\x00\xFF')
        self.assertEqual(encd, {'kind': 'ByteArray',
                                'type': ihm.format_bcif._Uint8})

        # type 5 (unsigned short)
        data, encd = d([256, 44033])
        self.assertEqual(data, b'\x00\x01\x01\xAC')
        self.assertEqual(encd, {'kind': 'ByteArray',
                                'type': ihm.format_bcif._Uint16})

        # type 6 (unsigned int)
        data, encd = d([4278255872])
        self.assertEqual(data, b'\x00\x01\x01\xFF')
        self.assertEqual(encd, {'kind': 'ByteArray',
                                'type': ihm.format_bcif._Uint32})

        # type 32 (32-bit float)
        data, encd = d([42.0])
        self.assertEqual(len(data), 4)
        self.assertEqual(encd, {'kind': 'ByteArray',
                                'type': ihm.format_bcif._Float32})

        # Too-large ints should cause an error
        self.assertRaises(TypeError, d, [2**34])
        self.assertRaises(TypeError, d, [-2**34])

    def test_delta_encoder(self):
        """Test Delta encoder"""
        d = ihm.format_bcif._DeltaEncoder()

        # too-small data is returned unchanged
        data = [0, 1, -1]
        encdata, encdict = d(data)
        self.assertEqual(data, encdata)
        self.assertIsNone(encdict)

        # large data is encoded
        data = [0, 1, -1] + [-1] * 40
        encdata, encdict = d(data)
        self.assertEqual(encdata, [0, 1, -2] + [0] * 40)
        self.assertEqual(encdict, {'origin': 0, 'kind': 'Delta',
                                   'srcType': ihm.format_bcif._Int8})

    def test_run_length_encoder(self):
        """Test RunLength encoder"""
        d = ihm.format_bcif._RunLengthEncoder()

        # too-small data is returned unchanged
        data = [0, 1, -1]
        encdata, encdict = d(data)
        self.assertEqual(data, encdata)
        self.assertIsNone(encdict)

        # large data that can't be compressed is returned unchanged
        data = list(range(50))
        encdata, encdict = d(data)
        self.assertEqual(data, encdata)
        self.assertIsNone(encdict)

        # large data that can be compressed
        data = [0] * 30 + [1] * 40
        encdata, encdict = d(data)
        self.assertEqual(encdata, [0, 30, 1, 40])
        self.assertEqual(encdict, {'kind': 'RunLength', 'srcSize': 70,
                                   'srcType': ihm.format_bcif._Uint8})

    def test_encode(self):
        """Test _encode function"""
        data = [1, 1, 1, 2, 3, 3]
        encoders = [ihm.format_bcif._ByteArrayEncoder()]
        encdata, encds = ihm.format_bcif._encode(data, encoders)
        self.assertEqual(encdata, b'\x01\x01\x01\x02\x03\x03')
        self.assertEqual(encds, [{'kind': 'ByteArray',
                                  'type': ihm.format_bcif._Uint8}])

        # DeltaEncoder will be a noop here since data is small
        encoders = [ihm.format_bcif._DeltaEncoder(),
                    ihm.format_bcif._ByteArrayEncoder()]
        encdata, encds = ihm.format_bcif._encode(data, encoders)
        self.assertEqual(encdata, b'\x01\x01\x01\x02\x03\x03')
        self.assertEqual(encds, [{'kind': 'ByteArray',
                                  'type': ihm.format_bcif._Uint8}])

    def test_mask_type_no_mask(self):
        """Test get_mask_and_type with no mask"""
        data = [1, 2, 3, 4]
        mask, typ = ihm.format_bcif._get_mask_and_type(data)
        self.assertIsNone(mask)
        self.assertEqual(typ, int)

    def test_mask_type_masked_int(self):
        """Test get_mask_and_type with masked int data"""
        data = [1, 2, 3, None, ihm.unknown, 4]
        mask, typ = ihm.format_bcif._get_mask_and_type(data)
        self.assertEqual(mask, [0, 0, 0, 1, 2, 0])
        self.assertEqual(typ, int)

    def test_mask_type_masked_float(self):
        """Test get_mask_and_type with masked float data"""
        data = [1.0, 2.0, 3.0, None, ihm.unknown, 4.0]
        mask, typ = ihm.format_bcif._get_mask_and_type(data)
        self.assertEqual(mask, [0, 0, 0, 1, 2, 0])
        self.assertEqual(typ, float)

    def test_mask_type_masked_numpy_float(self):
        """Test get_mask_and_type with masked numpy float data"""
        try:
            import numpy
        except ImportError:
            self.skipTest("this test requires numpy")
        data = [numpy.float64(4.2), None, ihm.unknown]
        mask, typ = ihm.format_bcif._get_mask_and_type(data)
        self.assertEqual(mask, [0, 1, 2])
        self.assertEqual(typ, float)

    def test_mask_type_masked_str(self):
        """Test get_mask_and_type with masked str data"""
        # Literal . and ? should not be masked
        data = ['a', 'b', None, ihm.unknown, 'c', '.', '?']
        mask, typ = ihm.format_bcif._get_mask_and_type(data)
        self.assertEqual(mask, [0, 0, 1, 2, 0, 0, 0])
        self.assertEqual(typ, str)

    def test_mask_type_mix_int_float(self):
        """Test get_mask_and_type with a mix of int and float data"""
        data = [1, 2, 3, 4.0]
        mask, typ = ihm.format_bcif._get_mask_and_type(data)
        self.assertIsNone(mask)
        self.assertEqual(typ, float)  # int/float is coerced to float

    def test_mask_type_mix_int_float_str(self):
        """Test get_mask_and_type with a mix of int/float/str data"""
        data = [1, 2, 3, 4.0, 'foo']
        mask, typ = ihm.format_bcif._get_mask_and_type(data)
        self.assertIsNone(mask)
        self.assertEqual(typ, str)  # int/float/str is coerced to str

    def test_mask_type_bad_type(self):
        """Test get_mask_and_type with unknown type data"""
        class MockObject:
            pass
        data = [MockObject()]
        self.assertRaises(ValueError, ihm.format_bcif._get_mask_and_type, data)

    def test_masked_encoder(self):
        """Test MaskedEncoder base class"""
        e = ihm.format_bcif._MaskedEncoder()
        e(None, None)  # noop

    def test_string_array_encoder_no_mask(self):
        """Test StringArray encoder with no mask"""
        d = ihm.format_bcif._StringArrayMaskedEncoder()
        indices, encs = d(['a', 'AB', 'a'], None)
        self.assertEqual(indices, b'\x00\x01\x00')
        enc, = encs
        self.assertEqual(enc['dataEncoding'],
                         [{'kind': 'ByteArray',
                           'type': ihm.format_bcif._Uint8}])
        self.assertEqual(enc['offsetEncoding'],
                         [{'kind': 'ByteArray',
                           'type': ihm.format_bcif._Uint8}])
        self.assertEqual(enc['offsets'], b'\x00\x01\x03')
        self.assertEqual(enc['stringData'], 'aAB')

    def test_string_array_encoder_mask(self):
        """Test StringArray encoder with mask"""
        d = ihm.format_bcif._StringArrayMaskedEncoder()
        # True should be mapped to 'YES'; int 3 to str '3'
        # Unmasked literal . and ? should be kept as-is
        indices, encs = d(['a', 'AB', True, ihm.unknown, None, 'a', 3,
                           '.', '?'],
                          [0, 0, 0, 2, 1, 0, 0, 0, 0])
        # \xff is -1 (masked value) as a signed char (Int8)
        self.assertEqual(indices, b'\x00\x01\x02\xff\xff\x00\x03\x04\x05')
        enc, = encs
        self.assertEqual(enc['dataEncoding'],
                         [{'kind': 'ByteArray',
                           'type': ihm.format_bcif._Int8}])
        self.assertEqual(enc['offsetEncoding'],
                         [{'kind': 'ByteArray',
                           'type': ihm.format_bcif._Uint8}])
        self.assertEqual(enc['offsets'], b'\x00\x01\x03\x06\x07\x08\t')
        self.assertEqual(enc['stringData'], 'aABYES3.?')

    def test_int_array_encoder_no_mask(self):
        """Test IntArray encoder with no mask"""
        d = ihm.format_bcif._IntArrayMaskedEncoder()
        data, encs = d([5, 7, 8], None)
        self.assertEqual(data, b'\x05\x07\x08')
        self.assertEqual(encs, [{'kind': 'ByteArray',
                                 'type': ihm.format_bcif._Uint8}])

    def test_int_array_encoder_mask(self):
        """Test IntArray encoder with mask"""
        d = ihm.format_bcif._IntArrayMaskedEncoder()
        data, encs = d([5, 7, '?', 8, None], [0, 0, 2, 0, 1])
        # \xff is -1 (masked value) as a signed char (Int8)
        self.assertEqual(data, b'\x05\x07\xff\x08\xff')
        self.assertEqual(encs, [{'kind': 'ByteArray',
                                 'type': ihm.format_bcif._Int8}])

    def test_float_array_encoder_no_mask(self):
        """Test FloatArray encoder with no mask"""
        d = ihm.format_bcif._FloatArrayMaskedEncoder()
        # int data should be coerced to float
        data, encs = d([5.0, 7.0, 8.0, 4], None)
        self.assertEqual(len(data), 4 * 4)
        self.assertEqual(encs, [{'kind': 'ByteArray',
                                 'type': ihm.format_bcif._Float32}])

    def test_float_array_encoder_mask(self):
        """Test FloatArray encoder with mask"""
        d = ihm.format_bcif._FloatArrayMaskedEncoder()
        data, encs = d([5., 7., '?', 8., None], [0, 0, 2, 0, 1])
        self.assertEqual(len(data), 5 * 4)
        self.assertEqual(encs, [{'kind': 'ByteArray',
                                 'type': ihm.format_bcif._Float32}])

    def test_category(self):
        """Test CategoryWriter class"""
        fh = MockFh()
        sys.modules['msgpack'] = MockMsgPack
        writer = ihm.format_bcif.BinaryCifWriter(fh)
        writer.start_block('ihm')
        with writer.category('foo') as loc:
            loc.write(bar='baz')
        writer.flush()
        block, = fh.data['dataBlocks']
        category, = block['categories']
        column, = category['columns']
        self.assertEqual(block['header'], 'ihm')
        self.assertEqual(category['name'], 'foo')
        self.assertEqual(category['rowCount'], 1)
        self.assertEqual(column['name'], 'bar')
        self.assertEqual(column['data']['encoding'][0]['stringData'],
                         'baz')

    def test_empty_loop(self):
        """Test LoopWriter class with no values"""
        fh = MockFh()
        sys.modules['msgpack'] = MockMsgPack
        writer = ihm.format_bcif.BinaryCifWriter(fh)
        writer.start_block('ihm')
        with writer.loop('foo', ["bar", "baz"]):
            pass
        writer.flush()
        self.assertEqual(fh.data['dataBlocks'][0]['categories'], [])

    def test_loop(self):
        """Test LoopWriter class"""
        fh = MockFh()
        sys.modules['msgpack'] = MockMsgPack
        writer = ihm.format_bcif.BinaryCifWriter(fh)
        writer.start_block('ihm')
        with writer.loop('foo', ["bar", "baz"]) as lp:
            lp.write(bar='x')
            lp.write(bar=None, baz='z')
            lp.write(bar=ihm.unknown, baz='z')
            lp.write(bar='.', baz='z')
            lp.write(bar='?', baz='z')
            lp.write(baz='y')
        writer.flush()
        block, = fh.data['dataBlocks']
        category, = block['categories']
        self.assertEqual(category['name'], 'foo')
        self.assertEqual(category['rowCount'], 6)
        cols = sorted(category['columns'], key=lambda x: x['name'])
        self.assertEqual(len(cols), 2)
        # Check mask for bar column; literal . and ? should not be masked (=0)
        self.assertEqual(cols[0]['mask']['data'],
                         b'\x00\x01\x02\x00\x00\x01')


if __name__ == '__main__':
    unittest.main()



import unittest
import ihm.reader
import ihm.dumper
import urllib.request
import os


class Tests(unittest.TestCase):
    def _read_cif(self, pdb_id):
        url = 'https://pdb-ihm.org/cif/%s.cif' % pdb_id
        with urllib.request.urlopen(url) as fh:
            s, = ihm.reader.read(fh)
        return s

    def _write_cif(self, s, check=True):
        with open('test.cif', 'w') as fh:
            ihm.dumper.write(fh, [s], check=check)
        os.unlink('test.cif')

    def test_9a0e(self):
        """Test IMP structure without errors (9a0e)"""
        s = self._read_cif('9a0e')
        # This used to have an incorrect reference sequence, now fixed.
        self._write_cif(s)

    def test_8zzd(self):
        """Test docking structure without errors (8zzd)"""
        s = self._read_cif('8zzd')
        # This used to have an incorrect assembly, now fixed.
        self._write_cif(s)

    def test_9a82(self):
        """Test HADDOCK structure without errors (9a82)"""
        s = self._read_cif('9a82')
        self._write_cif(s)

    def test_9a13(self):
        """Test HADDOCK structure without errors (9a13)"""
        s = self._read_cif('9a13')
        # This used to fail because feature #71 was null, but this
        # is now fixed.
        self._write_cif(s)

    def test_9a0t(self):
        """Test IMP structure without errors (9a0t)"""
        s = self._read_cif('9a0t')
        self._write_cif(s)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3

"""Check the output of each example for validity against the PDBx
   and IHM dictionaries.

   This should be periodically rechecked in case the PDBx and IHM dictionaries
   are updated.
"""

import sys
import os
import subprocess
import ihm.dictionary
import urllib.request

with urllib.request.urlopen(
        'http://mmcif.wwpdb.org/dictionaries/ascii/mmcif_pdbx_v50.dic') as fh:
    d_pdbx = ihm.dictionary.read(fh)
with urllib.request.urlopen(
        'http://mmcif.wwpdb.org/dictionaries/ascii/mmcif_ihm.dic') as fh:
    d_ihm = ihm.dictionary.read(fh)
pdbx_ihm = d_pdbx + d_ihm

for script in ('simple-docking.py', 'ligands_water.py',
               'non_standard_residues.py'):
    print(script)
    subprocess.check_call([sys.executable, '../examples/' + script])
    with open('output.cif') as fh:
        try:
            pdbx_ihm.validate(fh)
        except ihm.dictionary.ValidatorError as exc:
            print(exc)
os.unlink('output.cif')
